#!/usr/bin/perl -w

=head1 NAME

B<POWER::SQL> - Generate basic SQL queries

=head1 SYNOPSIS

 use POWER::SQL;

=head1 DESCRIPTION

=head2 

=over 4

=item *

       PRIMARY KEY.

=item *

   (JOIN)  ,   - 
    PRIMARY KEY     
.     .

=item *

   "_"   ,  
(         ).
    CGI  %Q    
,   "_".

=item *

    "-" .    
-order  -group .

=item *

   "__"     
   .

=item *

 Insert()  Update()    
_cdate (  )  _mdate (   )
  ,      .

=item *

    count(*),     
__count .

=back

=head1 METHODS

    POWER::SQL     DBI,  
      DBI: 

$row = $dbh->Row("SELECT * FROM Customers WHERE id_cust=?",555);

=over 4

=cut

package POWER::SQL;

use strict;
use vars qw($VERSION);

$VERSION = "1.00";

sub dbq { $_[0]->quote($_[2]) }
sub is_arr {ref $_[2] eq "ARRAY"}
sub is_int {$_[2] =~ m/^-?\d+ +(?:SECOND|MINUTE|HOUR|DAY|MONTH|YEAR)$/}
my %Func = (
    "eq"        => sub { "$_[1] =  ".&dbq },
    "ne"        => sub { "$_[1] != ".&dbq },
    "lt"        => sub { "$_[1] <  ".&dbq },
    "gt"        => sub { "$_[1] >  ".&dbq },
    "le"        => sub { "$_[1] <= ".&dbq },
    "ge"        => sub { "$_[1] >= ".&dbq },
    "btw"       => sub { &is_arr and "$_[1] BETWEEN ".dbq($_[0],"",$_[2][0])." AND ".dbq($_[0],"",$_[2][1]) },
    "date_eq"   => sub { &is_int and "$_[1] =  DATE_ADD(NOW(), INTERVAL $_[2])" },
    "date_ne"   => sub { &is_int and "$_[1] != DATE_ADD(NOW(), INTERVAL $_[2])" },
    "date_lt"   => sub { &is_int and "$_[1] <  DATE_ADD(NOW(), INTERVAL $_[2])" },
    "date_gt"   => sub { &is_int and "$_[1] >  DATE_ADD(NOW(), INTERVAL $_[2])" },
    "date_le"   => sub { &is_int and "$_[1] <= DATE_ADD(NOW(), INTERVAL $_[2])" },
    "date_ge"   => sub { &is_int and "$_[1] >= DATE_ADD(NOW(), INTERVAL $_[2])" },
    );
my %DESC;

=item B<P2SQL>( $table, \%P )

=item B<P2SQL>( \@tables, \%P )

 $table  @tables   ,    ,  %P
  ,       $table  @tables.
   ( {ID}, {Field}, ...)    "."
    .

   :

=over 4

=item {DB}

      

=item {Desc}

       
  "" => $dbh->All("DESC ")

=item {Table}

   

=item {INNER}

   INNER JOIN

=item {LEFT}

   LEFT JOIN

=item {ID}

  PRIMARY KEY  {Table}

=item {FieldList}

  "=, 2=2"  
%P   ""

=item {FieldAND}

a-la {FieldList},     "AND" 
 -    "__"
        "1"

=item {Update}

a-la {FieldAND},      "" 
 primary key,   .

=item {Order}

  " ASC, 2 DESC"  ""

=item {Group}

a-la {Order}

=back

=cut

sub DBI::db::P2SQL {
    my ($dbh, $tables, $P) = @_;
    my %SQL;
    #        
    #  $SQL{DB}
    $DESC{$dbh} ||= {map {$_=>undef} @{$dbh->selectcol_arrayref("SHOW TABLES")}};
    $SQL{DB} = [ keys %{$DESC{$dbh}} ];
    #      ,     
    my @tables = grep { exists $DESC{$dbh}{$_} } ref $tables ? @$tables : $tables;
    # ,      
    return %SQL unless @tables;
    #      
    for my $t (@tables) {
        unless ($DESC{$dbh}{$t}) {
            (my $sth = $dbh->prepare("DESC $t"))->execute();
            $DESC{$dbh}{$t} = $sth->fetchall_arrayref({});
        }
    }
    #  $SQL{Table}
    $SQL{Table} = $tables[0];
    #  $SQL{Desc}   
    $SQL{Desc}{ $SQL{Table} } = $DESC{$dbh}{ $SQL{Table} };
    #  $SQL{ID}
    $SQL{ID} = $SQL{Desc}{ $SQL{Table} }[0]{Field};
    #  ,     
    my @join = $tables[0];
    my @field = map {{ map {$_->{Field}=>1} @{$DESC{$dbh}{$_}} }} @tables;
    for my $right (1..$#tables) {
        my $rkey = $DESC{$dbh}{ $tables[$right] }[0]{Field};
        for my $left (0..$right-1) {
            my $lkey = $DESC{$dbh}{ $tables[$left] }[0]{Field};
            my $key = $field[$left]{$rkey}  ? $rkey : 
                      $field[$right]{$lkey} ? $lkey : next;
            push @join, "JOIN $tables[$right]".
                " ON ($tables[$right].$key=$tables[$left].$key)";
            #  $SQL{Desc}    
            $SQL{Desc}{ $tables[$right] } = $DESC{$dbh}{ $tables[$right] };
            last;
        }
    }
    #  $SQL{INNER}  $SQL{LEFT}
    $SQL{INNER} = join " INNER ", @join;
    $SQL{LEFT}  = join " LEFT ",  @join;
    #       
    my %qualify;
    for my $t (@tables) {
        next unless exists $SQL{Desc}{$t}; #     join
        $qualify{$_} ||= "$t.$_" for map {$_->{Field}} @{$SQL{Desc}{$t}};
    }
    #  $SQL{FieldList}, $SQL{FieldAND}, $SQL{Update}
    for my $f (keys %$P) {
        my ($name, $func) = $f =~ /^(.*?)(?:__(.*))?$/;
        next unless my $qname = $qualify{$name};            #   
        next unless my $sub   = $Func{$func || "eq"};       #   
        next unless my $expr  = $sub->($dbh, $qname, $P->{$f});# . .
        $SQL{FieldList} .= ", $expr" unless $func;
        $SQL{FieldAND}  .= " AND $expr";
        $SQL{Update}    .= " AND $expr" if $func or $name eq $SQL{ID};
    }
    $SQL{FieldList} =~ s/^, //;
    $SQL{FieldAND}  =~ s/^ AND //; $SQL{FieldAND} ||= "1";
    $SQL{Update}    =~ s/^ AND //; $SQL{Update}   ||= "1";
    #  $SQL{Order}
    for my $order (ref $P->{-order} ? @{$P->{-order}} : $P->{-order}) {
        next unless $order =~ /^\w+(?: ASC| DESC)$/;
        $SQL{Order} = join ", ", ($SQL{Order}||()), $order;
    }
    #  $SQL{Group}
    for my $group (ref $P->{-group} ? @{$P->{-group}} : $P->{-group}) {
        next unless $group =~ /^\w+(?: ASC| DESC)$/;
        $SQL{Group} = join ", ", ($SQL{Group}||()), $group;
    }
    return %SQL;
}

=item B<Insert>( $table, \%P )

INSERT INTO ?table SET ?set

     "_cdate",  "_mdate"     NOW().
 LAST_INSERT_ID()     undef().

=cut

sub DBI::db::Insert {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {"_cdate"=>undef, "_mdate"=>undef, %$P});
    return unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_[cm]date\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("INSERT INTO $SQL{Table} SET $SQL{FieldList}")
        ? $dbh->selectrow_array("SELECT LAST_INSERT_ID()") : undef;
}

=item B<Update>( $table, \%P )

UPDATE ?table SET ?set WHERE ?id=... AND ?where

     "_mdate",  '_mdate=NOW()'.
   SET     "",    WHERE  
  "__"  primary_key     "".
    UPDATE  undef().

=cut

sub DBI::db::Update {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {"_mdate"=>undef, %$P});
    return unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_mdate\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("UPDATE $SQL{Table} SET $SQL{FieldList} WHERE $SQL{Update}");
}

=item B<Replace>( $table, \%P )

REPLACE ?table SET ?set

     "_cdate"  "_mdate",     NOW().
   SET     "".
    REPLACE  undef().

=cut

sub DBI::db::Replace {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {"_cdate"=>undef, "_mdate"=>undef, %$P});
    return unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_[cm]date\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("REPLACE INTO $SQL{Table} SET $SQL{FieldList}");
}

=item B<Delete>( $tables, \%P )

DELETE FROM ?table WHERE ?where

    .  $tables  undef(),  
         %P .

=cut

sub DBI::db::Delete {
    my ($dbh, $table, $P) = @_;
    if (not defined $table) {
        return unless my @field = map {s/__.*//;$_} keys %$P;
        my %SQL = $dbh->P2SQL({$dbh->P2SQL()}->{DB});
        for my $t (@{$SQL{DB}}) {
            my %seen;
            push @$table, $t if @field == grep {++$seen{$_}==2} 
                @field, map {$_->{Field}} @{$SQL{Desc}{$t}};
        }
    }
    for my $t (ref $table ? @$table : $table) {
        my %SQL = $dbh->P2SQL($t, $P);
        next unless $SQL{Table};
        $dbh->do("DELETE FROM $SQL{Table} WHERE $SQL{FieldAND}");
    }
}

=item B<ID>( $tables, \%P )

SELECT primary_key FROM ?inner WHERE ?where [ORDER BY ?order]

   Col() .

=cut

sub DBI::db::ID {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    return unless $SQL{Table};
    $dbh->Col("SELECT $SQL{ID} FROM $SQL{INNER} WHERE $SQL{FieldAND}".
        ($SQL{Order} ? " ORDER BY $SQL{Order}" : ""));
}

=item B<Count>( $tables, \%P )

SELECT count(*) FROM ?inner WHERE ?where

   Col() .

=cut

sub DBI::db::Count {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    return unless $SQL{Table};
    $dbh->Col("SELECT count(*) __count FROM $SQL{INNER} WHERE $SQL{FieldAND}");
}

=item B<Select>( $tables, \%P )

SELECT * FROM ?inner WHERE ?where [ORDER BY ?order]
SELECT *, ?group, count(*) __count FROM ?inner WHERE ?where GROUP BY ?group
    [ORDER BY ?order]

   All()      Row()  .

=cut

sub DBI::db::Select {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    return unless $SQL{Table};
    my $SQL = "SELECT *".
        ($SQL{Group} ? ", $SQL{Group}, count(*) __count" : "").
        " FROM $SQL{INNER} WHERE $SQL{FieldAND}".
        ($SQL{Group} ? " GROUP BY $SQL{Group}" : "").
        ($SQL{Order} ? " ORDER BY $SQL{Order}" : "");
    return wantarray ? $dbh->All($SQL) : $dbh->Row($SQL);
}

=item B<All>($sql[, @bind])

Execute $sql and return array of hashrefs to row data.
Example:

 [ { id => 1, user => "root" },
   { id => 2, user => "daemon" },
   { id => 3, user => "system" },
 ]

=cut

sub DBI::db::All {
    my ($dbh, $sql, @bind) = @_;
    (my $sth = $dbh->prepare($sql))->execute(@bind);
    @{ $sth->fetchall_arrayref({}) };
}

=item B<Row>($sql[, @bind])

Execute $sql and return reference to hash with row data or undef.
Example:

 { id => 1, user => "root" }

=cut

sub DBI::db::Row {
    my ($dbh, $sql, @bind) = @_;
    (my $sth = $dbh->prepare($sql))->execute(@bind);
    $sth->fetchrow_hashref();
}

=item B<Col>($sql[, @bind])

Execute $sql and return array of column data in list context or
first cell in scalar context.

=cut

sub DBI::db::Col {
    my ($dbh, $sql, @bind) = @_;
    my @res = @{ $dbh->selectcol_arrayref($sql, undef, @bind) };
    return wantarray ? @res : $res[0];
}

=back

=head1 FUNCTIONS for FIELDS

,         "__":

=over 8

=item B<eq>, B<ne>, B<lt>, B<gt>, B<le>, B<ge>

name = != < > <= >= value

value must be /^\w+$/

=item B<btw>

name BETWEEN value[0] AND value[1]

value[0] and value[1] must be /^\w+$/

=item B<date_eq>, B<date_ne>, B<date_lt>, B<date_gt>, B<date_le>, B<date_ge>

name = != < > <= >= DATE_ADD(NOW(), INTERVAL value)

value must be /^-?\d+ (?:SECOND|MINUTE|HOUR|DAY|MONTH|YEAR)$/

=back

=head1 HISTORY

 1.00         15 16:52:30 EET 2002
        - first release

=head1 AUTHOR

Alex Efros <powerman@asdfGroup.com>

=cut

1;
